/**
 * @file Command.cpp
 * @author Copyright (c) 2013 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "Command.h"

#include <Common/era_api/err_codes.h>
#include "autocomplete/autocomplete.h"
#include "commands/Command_RunScript.h"
#include "util/stringUtils.h"
#include "Element.h"
#include "EracmdExceptions.h"
#include "Token.h"

Command *Command::AddKeyword(const std::wstring &text, Command::SyntaxType type, int skipArgs)
{
	return AddSyntax(text, -1, type, 0, NULL, skipArgs);
}

Command *Command::AddParam(unsigned int argNum, Command::SyntaxType type, unsigned int helpId, std::shared_ptr<ParamValidator> validator)
{
	return AddSyntax(L"", argNum, type, helpId, validator, 0);
}

Command *Command::AddSyntax(const std::wstring &text, unsigned int argNum, Command::SyntaxType syntaxType, unsigned int helpId, std::shared_ptr<ParamValidator> validator, int skipArgs = 0)
{
	if (syntaxType == conditional && (!syntax.empty()) && syntax.rbegin()->syntaxType == mandatory)
		throw std::exception("error: conditional syntax can not follow mandatory syntax");

	if ((text.empty() && argNum == -1) || ((!text.empty()) && argNum != -1))
		throw std::exception("error: only TEXT or only ARGNUM should be defined for syntax");

	if (text.empty() && argNum>20)
		throw std::exception("error: ARGNUM can not exceed 20");

	CommandSyntax s;
	s.text = text;
	s.argNum = argNum;
	s.syntaxType = syntaxType;
	s.helpId = helpId;
	s.validator = validator;
	s.autoCompleteSkipArgs = skipArgs;
	syntax.push_back(s);
	return this;
}

void Command::Execute(elementArray &elements, OutputHandler &out)
{

}

std::shared_ptr<Element> Command::GetArgument(unsigned int argNum)
{
	if (argNum >= arguments.size())
		return std::shared_ptr<Element>(NULL);
	else
		return arguments[argNum];
}

std::wstring Command::GetArgumentAsString(unsigned int argNum)
{
	std::shared_ptr<Element> el = GetArgument(argNum);
	if (el == NULL) return L"";
	return el->GetString();
}


std::wstring FormatCommand(elementArray &elements, const std::wstring &commandString)
{
	std::wstring line = ElementArrayToString(elements);
	if (line.empty())
		line = commandString;
	else
		line = commandString + L" " + line;
	return line;
}


void Command::GrabArguments(elementArray &elements, const std::wstring &commandString, AutocompleteCollector *collector)
{
	arguments.clear();
	bool skipConditional = false;
	elementArray::iterator elementPos = elements.begin();
	elementArray::iterator elementEnd = elements.end();

	std::shared_ptr<Element> lastElement = elements.size() ? *elements.rbegin() : NULL;
	std::wstring lastElementString = elements.size() ? lastElement->GetString() : L"";
	
	std::set<size_t> optionalBlocksUsed;

	int syntaxStart = -1;
	bool finish=false;
	bool mandatoryBlockFinished = false;
	do
	{
		finish = elementPos == elements.end();
		bool sequenceMatched = false;

		for (size_t i = (syntaxStart==-1)?0:syntaxStart; i < syntax.size(); ++i)
		{
			if (syntax[i].syntaxType != conditional)
				skipConditional = false;
			else
				if (skipConditional)
					continue;

			if (syntax[i].syntaxType==mandatory && mandatoryBlockFinished)
			{
				//no optional parameters, we are done
				break;
			}

			if (syntax[i].syntaxType==optional)
			{
				if (syntaxStart==-1)
				{
					syntaxStart=(int)i; //only optional blocks should follow, note that first optional is processed twice
				}
				else 
				{
					if (sequenceMatched)
						break; //restart new optional block

					if (optionalBlocksUsed.find(i)!=optionalBlocksUsed.end())
					{
						skipConditional=true;
						continue;	
					}
				}
			}

			bool isMatching = false;
			if (elementPos != elements.end()) // i can check it
			{
				if (syntax[i].text.empty() || _wcsicmp(syntax[i].text.c_str(), (*elementPos)->GetString().c_str()) == 0) // argument or matching string
				{
					isMatching = true;
				}
			}
			if (!isMatching)
			{				
				if (collector) // autocomplete
				{
					if (collector->openWord.empty())
					{ // there is no last word
						if (elementPos == elements.end() && (lastElement == NULL || (lastElement->openedList == false && lastElement->missingOperand == false)))
						{
							if (syntax[i].validator)
							{
								syntax[i].validator->AutoComplete(NULL, this, collector);
							}
							else
							{
								int curArg = (int) arguments.size();
								if (curArg >= syntax[i].autoCompleteSkipArgs)
									collector->Add(syntax[i].text);
							}
						}
					}
					else // last word is prefix of this syntax word
						if (elementPos != elements.end() && (*elementPos == lastElement) && (lastElementString.length() <= syntax[i].text.length()) && (_wcsnicmp(lastElementString.c_str(), syntax[i].text.c_str(), lastElementString.length()) == 0))
						{
							collector->Add(syntax[i].text);
						}
				}

				
				if (syntax[i].syntaxType == optional) // optional -> continue
				{
					skipConditional = true;
					continue;
				}
				else // error
				{
					// get expected param name
					std::wstring paramName;
					if (syntax[i].helpId>0)
					{
						paramName = LoadStringFromResource(syntax[i].helpId);
						size_t pipeLocation = paramName.find(L'~');
						if (pipeLocation == std::wstring::npos)
							throw std::exception(("missing ~ in argument help for command " + WstringToString(commandString)).c_str());
						paramName.erase(paramName.begin() + pipeLocation, paramName.end());
					}
					else paramName = syntax[i].text;


					if (elementPos != elements.end())
					{
						throw EraCMDException(CMDERROR_COMMAND_UNEXPECTED_ARGUMENT, (*elementPos)->GetString().c_str(), paramName.c_str());
					}
					else
					{
						throw EraCMDException(CMDERROR_COMMAND_MISSING_ARGUMENT, paramName.c_str());
					}
				}
			}
			else // matching
			{
				if (syntax[i].syntaxType==optional) 
					optionalBlocksUsed.insert(i);

				sequenceMatched = true;
				if (syntax[i].text.empty()) // take it as argument
				{
					unsigned int argNum = syntax[i].argNum;
					while (arguments.size() <= argNum)
						arguments.push_back(std::shared_ptr<Element>(NULL));

					arguments[argNum] = *elementPos;

					if (syntax[i].validator)
					{
						if (collector)
						{
							if (*elementPos == lastElement)
							{
								if (!collector->openWord.empty() && lastElement->openedList == false)
								{
									syntax[i].validator->AutoComplete(lastElement, this, collector);
								}
								else
								if (collector->openWord.empty() && (lastElement->openedList == true || lastElement->missingOperand == true))
								{
									syntax[i].validator->AutoComplete(lastElement, this, collector);
								}
							}
						}
						else syntax[i].validator->ValidateParam(*elementPos, this);
					}
				}
			}
			elementPos++;
		}
		mandatoryBlockFinished = true;
		if (!sequenceMatched && !finish)
		{
			if (elementPos==elements.begin() && (syntax.empty() || syntax[0].syntaxType==optional))
				break;

			throw EraCMDException(CMDERROR_COMMAND_UNEXPECTED_ARGUMENT2, (*elementPos)->GetString().c_str());
		}		
	} while (finish==false);

	if (elementPos != elements.end())
		throw EraCMDException(CMDERROR_COMMAND_TOO_MANY_ARGUMENTS, (*elementPos)->GetString().c_str());
}

