/**
 * @file ConnectionParams.h
 * @author Copyright (c) 2013 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "ConnectionParams.h"

#include <Common/era_api/err_codes.h>
#include "util/stringUtils.h"
#include "EracmdExceptions.h"
#include "resource.h"
#include "Token.h"

ConnectionParams::ConnectionParams()
	: host(CONN_DEFAULT_SERVER),
	port(CONN_DEFAULT_PORT),
	user(LoadStringFromResource(IDS_USERTYPE_ADM)),
	password(CONN_DEFAULT_PASSWORD),
	userDomain(false),
	userCurrent(false),
	askPassword(false),
	askAll(false)
{
}

bool ConnectionParamsProcessor::IsConnectionParam(const std::wstring &currTokenText)
{
	size_t prefixLen = wcslen(CONN_PAR_PREFIX);
	return currTokenText.length() >= prefixLen && currTokenText.substr(0, prefixLen) == CONN_PAR_PREFIX;
}

const std::wstring &ConnectionParamsProcessor::GetConnectionArgument(tokenArray::const_iterator tokensIt)
{
	if (tokensIt >= tokens.end())
		throw EraCMDException(CMDERROR_UNEXPECTED_INPUT_END);

	const std::wstring &currTokenText = (*tokensIt)->text;
	if (IsConnectionParam(currTokenText))
		throw EraCMDException(CMDERROR_UNEXPECTED_CONN_PARAM, currTokenText.c_str());
	return currTokenText;
}

void ConnectionParamsProcessor::GetUser(tokenArray::const_iterator tokensIt, ConnectionParams &connectionParams)
{
	connectionParams.user = GetConnectionArgument(tokensIt);

	if (connectionParams.user.empty())
		throw EraCMDException(CMDERROR_USERNAME_CANNOT_BE_EMPTY);

	std::wstring::size_type bSlashPos = connectionParams.user.find(L'\\');
	if (bSlashPos != std::wstring::npos)
		connectionParams.userDomain = true;
}


std::pair<tokenArray::const_iterator, tokenArray::const_iterator> ConnectionParamsProcessor::ProcessTo(ConnectionParams &connectionParams)
{
	tokenArray::const_iterator tokensIt;
	tokenArray::const_iterator firstConnTokenIt = tokens.end();
	struct ConnectionParamsSet
	{
		ConnectionParamsSet() : serverSet(false), userSet(false), userDomainSet(false), userCurrentSet(false), passwordSet(false), askPasswordSet(false), askAllSet(false) {}
		bool serverSet, userSet, userDomainSet, userCurrentSet, passwordSet, askPasswordSet, askAllSet;
	} connectionParamsSet;

	for (tokensIt = tokens.begin(); tokensIt != tokens.end(); ++tokensIt)
	{
		std::wstring currTokenText = (*tokensIt)->text;
		if (IsConnectionParam(currTokenText))
		{
			if (firstConnTokenIt == tokens.end())
				firstConnTokenIt = tokensIt;

			size_t prefixLen = wcslen(CONN_PAR_PREFIX);
			currTokenText = currTokenText.substr(prefixLen, currTokenText.length() - prefixLen);

			if (currTokenText == CONN_PAR_SERVER)
			{
				connectionParamsSet.serverSet = true;
				std::wstring currOption;
				currOption = GetConnectionArgument(++tokensIt);

				int ret = ParseHost(currOption, connectionParams.host, connectionParams.port);
				if (ret != 0)
					throw EraCMDException(CMDERROR_INVALID_PORT, currOption.c_str());
			}
			else if (currTokenText == CONN_PAR_USERNAME)
			{
				connectionParamsSet.userSet = true;
				GetUser(++tokensIt, connectionParams);
			}
			else if (currTokenText == CONN_PAR_USERNAME_DOMAIN)
			{
				connectionParamsSet.userDomainSet = true;
				GetUser(++tokensIt, connectionParams);
				connectionParams.userDomain = true;
			}
			else if (currTokenText == CONN_PAR_USERNAME_CURRENT)
			{
				connectionParamsSet.userCurrentSet = true;
				connectionParams.userCurrent = true;
				connectionParams.userDomain = true;
			}
			else if (currTokenText == CONN_PAR_PASSWORD)
			{
				connectionParamsSet.passwordSet = true;
				connectionParams.password = GetConnectionArgument(++tokensIt);
			}
			else if (currTokenText == CONN_PAR_PASSWORD_ASK)
			{
				connectionParamsSet.askPasswordSet = true;
				connectionParams.askPassword = true;
			}
			else if (currTokenText == CONN_PAR_ASK_ALL)
			{
				connectionParamsSet.askAllSet = true;
				connectionParams.askAll = true;
			}
			else if (currTokenText == CONN_PAR_ALTERNATIVE_STARTUP_SCRIPT)
			{
				connectionParams.alternativeStartupScript = GetConnectionArgument(++tokensIt);
			}
			else
			{
				if (currTokenText.empty())
					throw EraCMDException(CMDERROR_EMPTY_CONN_PARAM);
				else
					throw EraCMDException(CMDERROR_INVALID_CONN_PARAM, currTokenText.c_str());
			}
		}
		else
		{
			// did not find any more connection parameters, our work is done
			break;
		}
	}

	if (connectionParamsSet.askAllSet && (connectionParamsSet.serverSet || connectionParamsSet.userSet || connectionParamsSet.userDomainSet || connectionParamsSet.userCurrentSet || connectionParamsSet.passwordSet || connectionParamsSet.askPasswordSet))
		throw EraCMDException(CMDERROR_CONFLICT_OPTIONS_ASK_ALL);

	// check the command parameters conflicts
	if (connectionParamsSet.userSet && connectionParamsSet.userDomainSet)
		throw EraCMDException(CMDERROR_CONFLICT_OPTIONS_TWO, CONN_PAR_USERNAME, CONN_PAR_USERNAME_DOMAIN);
	if (connectionParamsSet.userCurrentSet && connectionParamsSet.userSet)
		throw EraCMDException(CMDERROR_CONFLICT_OPTIONS_TWO, CONN_PAR_USERNAME_CURRENT, CONN_PAR_USERNAME);
	if (connectionParamsSet.userCurrentSet && connectionParamsSet.userDomainSet)
		throw EraCMDException(CMDERROR_CONFLICT_OPTIONS_TWO, CONN_PAR_USERNAME_CURRENT, CONN_PAR_USERNAME_DOMAIN);
	if (connectionParamsSet.userCurrentSet && connectionParamsSet.passwordSet)
		throw EraCMDException(CMDERROR_CONFLICT_OPTIONS_TWO, CONN_PAR_USERNAME_CURRENT, CONN_PAR_PASSWORD);
	if (connectionParamsSet.userCurrentSet && connectionParamsSet.askPasswordSet)
		throw EraCMDException(CMDERROR_CONFLICT_OPTIONS_TWO, CONN_PAR_USERNAME_CURRENT, CONN_PAR_PASSWORD_ASK);
	if (connectionParamsSet.askPasswordSet && connectionParamsSet.passwordSet)
		throw EraCMDException(CMDERROR_CONFLICT_OPTIONS_TWO, CONN_PAR_PASSWORD_ASK, CONN_PAR_PASSWORD);

	return std::make_pair(firstConnTokenIt, tokensIt);
}
