/**
 * @file ConsoleKeyboardInput.cpp
 * @author Copyright (c) 2014 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "ConsoleKeyboardInput.h"

#include <Common/scope_exit.h>
#include "ConsoleHistory.h"
#include "ScriptInterpreter.h"

ConsoleKeyboardInput::ConsoleKeyboardInput(ScriptInterpreter &scriptInterpreter, bool inkognito) : autoComplete(scriptInterpreter)
{
	hStdin = GetStdHandle(STD_INPUT_HANDLE);
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);

	CONSOLE_CURSOR_INFO cursorInfo;
	GetConsoleCursorInfo(hStdout, &cursorInfo);
	normalCursorSize = cursorInfo.dwSize;
	overwrite = false;

	DWORD mode;
	GetConsoleMode(hStdin, &mode);
	mode |= ENABLE_WINDOW_INPUT;
	SetConsoleMode(hStdin, mode);

	historyActive = false;

	inkognitoMode = inkognito;
}

void ConsoleKeyboardInput::RedrawInput()
{
	CONSOLE_SCREEN_BUFFER_INFO info;
	GetConsoleScreenBufferInfo(hStdout, &info);

	int lineStartOffset = lineStartPosition.Y;

	SHORT oldNumLines = (SHORT)ceil((line.length() + lineStartOffset) / (double)consoleBufferSize.X);
	if (info.dwSize.Y <= lineStartPosition.Y)
	{
		lineStartPosition.Y = info.dwSize.Y - oldNumLines;
	}
	consoleBufferSize = info.dwSize;

	DWORD charsWritten;
	WriteText(hStdout, line.c_str(), (DWORD) line.length(), lineStartPosition, &charsWritten);
	SetConsoleCursorPosition(hStdout, GetCoordForPosition(cursorPos));

	WriteChar(hStdout, L' ', oldNumLines * consoleBufferSize.X - (DWORD) line.length() - lineStartOffset, GetCoordForPosition(cursorPos), &charsWritten);
}


void ConsoleKeyboardInput::GetInput()
{
	overwrite = true;
	ToggleInsert();

	line.clear();
	cursorPos = 0;

	CONSOLE_SCREEN_BUFFER_INFO info;
	GetConsoleScreenBufferInfo(hStdout, &info);
	lineStartPosition = info.dwCursorPosition;
	consoleBufferSize = info.dwSize;

	autoComplete.Reset();
	scope_exit resetAC(std::bind(&Autocomplete::Reset, &autoComplete));



	while (1)
	{
		DWORD cNumRead, i;
		INPUT_RECORD irInBuf[128];
		ReadConsoleInput(hStdin, irInBuf, 128, &cNumRead);
		for (i = 0; i < cNumRead; ++i)
		{
			switch(irInBuf[i].EventType)
			{
				case WINDOW_BUFFER_SIZE_EVENT:
				{
					RedrawInput();
					break;
				}
				case KEY_EVENT:
				{
					KEY_EVENT_RECORD ker = irInBuf[i].Event.KeyEvent;
					if (ker.bKeyDown)
					{
						if (ker.uChar.UnicodeChar >= L' ')
						{
							AddTypedChar(ker.uChar.UnicodeChar);
						}
						else
						switch (ker.wVirtualKeyCode)
						{
							case VK_RETURN:
								if (!inkognitoMode)
									gCmdHistory.Add(line);
								return; // finished
						// autocomplete
							case VK_TAB:
								AutoComplete(false);
								break;
							case VK_ESCAPE:
								if (!autoComplete.IsReset())
									AutoComplete(true);
								else
									ReplaceInput(historyActive ? lineB4History : L"");
								break;
						// move cursor
							case VK_RIGHT:
								MoveCursor(1);
								break;
							case VK_LEFT:
								MoveCursor(-1);
								break;
							case VK_END:
								MoveCursor(int(line.length()));
								break;
							case VK_HOME:
								MoveCursor(-int(line.length()));
								break;
						// history
							case VK_UP:
								GetHistory(true);
								break;
							case VK_DOWN:
								GetHistory(false);
								break;
						// insert
							case VK_INSERT:
								ToggleInsert();
								break;
						// delete
							case VK_DELETE:
								DeleteChar(cursorPos);
								break;
						// backspace
							case VK_BACK:
								if (cursorPos>0)
								{
									MoveCursor(-1);
									DeleteChar(cursorPos);
								}
								break;
							default:
								break;
						} // switch keycode

						if (ker.wVirtualKeyCode != VK_TAB)
							autoComplete.Reset();

						if (ker.wVirtualKeyCode != VK_UP && ker.wVirtualKeyCode != VK_DOWN)
							historyActive = false;

					} // if key down
					break;
				} // case key_event
				default:
					break;
			} // switch event type
		}
	}
}

void ConsoleKeyboardInput::AddTypedChar(WCHAR ch)
{
	if (overwrite && cursorPos<(int)line.length())
	{
		line.replace((size_t)cursorPos, 1, 1, ch);
		DWORD charsWritten;
		WriteChar(hStdout, ch, 1, GetCoordForPosition(cursorPos), &charsWritten);
	}
	else
	{
		line.insert((size_t)cursorPos, 1, ch);
		int writeLen = (int)line.length() - cursorPos;
		if (writeLen>0)
		{
			COORD posPos = GetCoordForPosition(cursorPos);
			DWORD charsWritten = 0;
			WriteText(hStdout, line.c_str() + cursorPos, writeLen, posPos, &charsWritten);
		}
	}
	MoveCursor(1);
}

void ConsoleKeyboardInput::DeleteChar(int pos)
{
	if (pos<0 || pos >= (int)line.length()) return;

	line.erase(pos, 1);
	COORD endPos = GetCoordForPosition((int)line.length());
	DWORD charsWritten = 0;
	WriteChar(hStdout, L' ', 1, endPos, &charsWritten);

	int writeLen = (int)line.length() - pos;
	if (writeLen>0)
	{
		COORD posPos = GetCoordForPosition(pos);
		DWORD charsWritten = 0;
		WriteText(hStdout, line.c_str() + pos, writeLen, posPos, &charsWritten);
	}
}

void ConsoleKeyboardInput::MoveCursor(int offset)
{
	cursorPos += offset;
	if (cursorPos<0) cursorPos = 0;
	if (cursorPos>(int)line.length()) cursorPos = (int)line.length();
	COORD pos = GetCoordForPosition(cursorPos);
	SetConsoleCursorPosition(hStdout, pos);
}

COORD ConsoleKeyboardInput::GetCoordForPosition(int pos)
{
	COORD result = lineStartPosition;
	result.X += pos;
	while (result.X >= consoleBufferSize.X)
	{
		result.X -= consoleBufferSize.X;
		result.Y++;
	}
	return result;
}

void ConsoleKeyboardInput::ToggleInsert()
{
	overwrite = !overwrite;
	CONSOLE_CURSOR_INFO cursorInfo;
	cursorInfo.bVisible = TRUE;
	cursorInfo.dwSize = overwrite ? 100 : normalCursorSize;
	SetConsoleCursorInfo(hStdout, &cursorInfo);
}

void ConsoleKeyboardInput::AutoComplete(bool remove)
{
	if (inkognitoMode)
		return;

	if (autoComplete.IsReset())
	{
		if (!remove)
			MoveCursor(int(line.length()));

		cursorAutoCompleteStartPosition = GetCoordForPosition(cursorPos);
		lastAutoCompleteSuggestion.clear();
		if (remove)
			return;
	}

	std::wstring completion;
	int keepChars = autoComplete.GetNext(line, completion);
	if (!completion.empty())
	{
		if (remove)
			completion.clear();
		int charsToDelete = (int)line.length() - keepChars;
		line.erase(keepChars, charsToDelete);
		line += completion;
		DWORD charsWritten;
		WriteText(hStdout, completion.c_str(), (int)completion.length(), cursorAutoCompleteStartPosition, &charsWritten);

		int numWhiteSpaces = int(lastAutoCompleteSuggestion.length() - completion.length());
		MoveCursor(-numWhiteSpaces);
		if (numWhiteSpaces>0)
		{
			CONSOLE_SCREEN_BUFFER_INFO info;
			GetConsoleScreenBufferInfo(hStdout, &info);
			DWORD charsWritten = 0;
			WriteChar(hStdout, L' ', numWhiteSpaces, info.dwCursorPosition, &charsWritten);
		}
		lastAutoCompleteSuggestion = completion;
	}
}

void ConsoleKeyboardInput::ReplaceInput(const std::wstring &newInput)
{
	int charsToDelete = int(line.length() - newInput.length());
	DWORD charsWritten = 0;

	if (charsToDelete>0)
	{
		COORD endPos = GetCoordForPosition((int)line.length() - charsToDelete);
		WriteChar(hStdout, L' ', charsToDelete, endPos, &charsWritten);
	}
	line = newInput;
	cursorPos = (int)line.length();
	COORD pos = GetCoordForPosition(cursorPos);
	SetConsoleCursorPosition(hStdout, pos);

	WriteText(hStdout, line.c_str(), (DWORD) line.length(), lineStartPosition, &charsWritten);
}

void ConsoleKeyboardInput::GetHistory(bool prior)
{
	if (inkognitoMode)
		return;

	if (historyActive == false)
	{
		historyActive = true;
		lineB4History = line;
	}

	std::wstring s = gCmdHistory.GetNextItem(prior);
	ReplaceInput(s);
}

void ConsoleKeyboardInput::WriteText(HANDLE hConsoleOutput, LPCWSTR lpCharacter, DWORD nLength, COORD dwWriteCoord, LPDWORD lpNumberOfCharsWritten)
{
	if (inkognitoMode == false)
	{
		WriteConsoleOutputCharacter(hConsoleOutput, lpCharacter, nLength, dwWriteCoord, lpNumberOfCharsWritten);
		ApplyCurrentAttributes(hConsoleOutput, *lpNumberOfCharsWritten, dwWriteCoord);
	}
	else
	{
		WriteChar(hConsoleOutput, L'*', nLength, dwWriteCoord, lpNumberOfCharsWritten);
	}
}

void ConsoleKeyboardInput::WriteChar(HANDLE hConsoleOutput, WCHAR cCharacter, DWORD nLength, COORD dwWriteCoord, LPDWORD lpNumberOfCharsWritten)
{
	if (cCharacter == L' ' || inkognitoMode == false)
		FillConsoleOutputCharacter(hConsoleOutput, cCharacter, nLength, dwWriteCoord, lpNumberOfCharsWritten);
	else
		FillConsoleOutputCharacter(hConsoleOutput, L'*', nLength, dwWriteCoord, lpNumberOfCharsWritten);

	ApplyCurrentAttributes(hConsoleOutput, *lpNumberOfCharsWritten, dwWriteCoord);
}

void ConsoleKeyboardInput::ApplyCurrentAttributes(HANDLE hConsoleOutput, DWORD nLength, COORD dwWriteCoord)
{
	CONSOLE_SCREEN_BUFFER_INFO info;
	if ((::GetConsoleScreenBufferInfo(hConsoleOutput, &info) != 0))
	{
		DWORD lpNumberOfAttrsWritten_ignored;
		FillConsoleOutputAttribute(hConsoleOutput, info.wAttributes, nLength, dwWriteCoord, &lpNumberOfAttrsWritten_ignored);
	}
}

