/**
 * @file PolicyTree.cpp
 * @author Copyright (c) 2013 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "PolicyTree.h"

#include <Common/PolicyConstants.h>

PolicyTreeInfo::PolicyTreeInfo()
	: PolicyBaseInfo(), nParentIndex(POLICY_DATA_ARRAY_DEFAULT_NOTEXISTING_INDEX)
{}
PolicyTreeInfo::PolicyTreeInfo(int nId, const std::wstring &policyId, const std::wstring &policyName, const std::wstring &policyPrimaryServer, DWORD dwFlags, int nPrimaryId, int nParentPrimaryId, const std::wstring &parentPrimaryServer, const std::wstring &repDirectParent)
	: PolicyBaseInfo(nId, policyId, policyName, policyPrimaryServer, dwFlags, nPrimaryId, nParentPrimaryId, parentPrimaryServer, repDirectParent), nParentIndex(POLICY_DATA_ARRAY_DEFAULT_NOTEXISTING_INDEX)
{}
PolicyTreeInfo::PolicyTreeInfo(int nId, const wchar_t *policyId, const wchar_t *policyName, const wchar_t *policyPrimaryServer, DWORD dwFlags, int nPrimaryId, int nParentPrimaryId, const wchar_t* parentPrimaryServer, const wchar_t* repDirectParent)
	: PolicyBaseInfo(nId, policyId, policyName, policyPrimaryServer, dwFlags, nPrimaryId, nParentPrimaryId, parentPrimaryServer, repDirectParent), nParentIndex(POLICY_DATA_ARRAY_DEFAULT_NOTEXISTING_INDEX)
{}

void PolicyTreeUtils::PrintPolicyTree(PolicyTreePtr parent, std::wstring namePrefix, std::vector<StringVector> &rows, bool printFirstRoot)
{
	if (parent == NULL)
		return;

	std::vector<PolicyTreePtr>::iterator childBeginIter;
	if (printFirstRoot)
	{
		StringVector row(2);
		row[0] = parent->data.policyId;
		row[1] = namePrefix + parent->data.policyName;
		rows.push_back(std::move(row));
		childBeginIter = parent->children.begin();
	}
	else
		childBeginIter = parent->children.size() > 0 ? parent->children.begin() + 1 : parent->children.begin();

	std::sort(
		childBeginIter,
		parent->children.end(), [](PolicyTreePtr ti1, PolicyTreePtr ti2) -> bool
		{
			int comp = _wcsicmp(ti1->data.policyName.c_str(), ti2->data.policyName.c_str());
			//if (comp == 0)
			//	return ti1->data.nId < ti2->data.nId;
			//else
				return comp < 0;
		}
	);

	std::for_each(
		parent->children.begin(),
		parent->children.end(),
		[printFirstRoot, &namePrefix, &rows](PolicyTreePtr &ti)
		{
			#define TREE_BRANCH_STR L"|__ "
			#define TREE_BRANCH_INDENT L"    "

			if (printFirstRoot)
				PrintPolicyTree(ti, namePrefix == L"" ? TREE_BRANCH_STR : TREE_BRANCH_INDENT + namePrefix, rows, true);
			else
				PrintPolicyTree(ti, namePrefix, rows, true);
		}
	);
}

PolicyTreePtr PolicyTreeUtils::AttachToTree(std::vector<PolicyTreeInfo> &policies, PolicyTreePtr parent, size_t nArrayChildIndex, const wchar_t *onlyPolicyOfServer, const wchar_t *primaryServer)
{
	if (nArrayChildIndex >= policies.size() || nArrayChildIndex < 0)
		return NULL;

	if (policies[nArrayChildIndex].nPrimaryId == POLICY_DEFAULT_PARENT_POLICY_ID || policies[nArrayChildIndex].nPrimaryId == POLICY_DEFAULT_CLIENT_POLICY_ID)
	{
		std::wstring str = L"";

		if (policies[nArrayChildIndex].nPrimaryId == POLICY_DEFAULT_PARENT_POLICY_ID)
		{
			for (size_t i = 0; i < policies.size(); ++i)
			{
				if ((policies[i].dwPolicyFlags & (POLICY_FLAG_ACTIVE_FLAG | POLICY_FLAG_REP_FROM_UPPER | POLICY_FLAG_DEFAULT_PARENT_FOR_SLAVE)) == (POLICY_FLAG_ACTIVE_FLAG | POLICY_FLAG_REP_FROM_UPPER | POLICY_FLAG_DEFAULT_PARENT_FOR_SLAVE))
				{
					str = policies[i].policyName;
					break;
				}
			}
		}
		else if (policies[nArrayChildIndex].nPrimaryId == POLICY_DEFAULT_CLIENT_POLICY_ID)
		{
			for (size_t i = 0; i < policies.size(); ++i)
			{
				if ((policies[i].dwPolicyFlags & POLICY_FLAG_DEFAULT_PARENT_FOR_CLIENTS) && _wcsicmp(primaryServer, policies[i].policyPrimaryServer.c_str()) == 0)
				{
					str = policies[i].policyName;
					break;
				}
			}
		}

		if (!str.empty())
			policies[nArrayChildIndex].policyName += L"  -->  " + str;
	}

	PolicyTreePtr item(new PolicyTree(policies[nArrayChildIndex]));
	if (parent == NULL)
		parent = item;
	else
		parent->AddChild(item);

	for (size_t i = 0; i < policies.size(); ++i)
	{
		if (
			(policies[i].nParentIndex == nArrayChildIndex) &&
			(i != nArrayChildIndex) &&
			(onlyPolicyOfServer && (_wcsicmp(policies[i].policyPrimaryServer.c_str(), onlyPolicyOfServer) == 0) || (onlyPolicyOfServer == NULL))
			)
		{
			AttachToTree(policies, item, i, onlyPolicyOfServer, primaryServer);
		}
	}

	return item;
}

void PolicyTreeUtils::BuildPolicyTree(std::vector<PolicyTreeInfo> &policies, PolicyTreePtr &rootItem, const wchar_t *onlyPolicyOfServer, const wchar_t *primaryServer)
{
	for (size_t i = 0; i < policies.size(); ++i)
	{
		for (size_t j = 0; j < policies.size(); ++j)
		{
			if (policies[j].nPrimaryId == policies[i].nParentPrimaryId && _wcsicmp(policies[j].policyPrimaryServer.c_str(), policies[i].parentPrimaryServer.c_str()) == 0 ||
				((policies[j].dwPolicyFlags & POLICY_FLAG_DEFAULT_PARENT_FOR_SLAVE) && policies[i].nParentPrimaryId == POLICY_DEFAULT_PARENT_POLICY_ID && _wcsicmp(policies[i].repDirectParent.c_str(), policies[j].policyPrimaryServer.c_str()) == 0))
			{
				policies[i].nParentIndex = (int) j;
				break;
			}
		}

		for (size_t j = 0; j < policies.size(); ++j)
		{
			if (policies[i].nParentPrimaryId != POLICY_DEFAULT_NONE_POLICY_ID && policies[j].nParentIndex == POLICY_DATA_ARRAY_DEFAULT_NOTEXISTING_INDEX &&
				(policies[j].nParentPrimaryId == policies[i].nParentPrimaryId  && _wcsicmp(policies[j].parentPrimaryServer.c_str(), primaryServer) == 0 || 
				((policies[i].dwPolicyFlags & POLICY_FLAG_DEFAULT_PARENT_FOR_SLAVE) && policies[j].nParentPrimaryId == POLICY_DEFAULT_PARENT_POLICY_ID && _wcsicmp(policies[j].repDirectParent.c_str(), primaryServer) == 0)))
				policies[j].nParentIndex = (int) i;
		}
	}

	for (size_t i = 0; i < policies.size(); ++i)
	{
		if (policies[i].dwPolicyFlags & POLICY_FLAG_VIRTUAL_PARENT_FOR_CLIENTS)
			continue;

		if (policies[i].nParentIndex == POLICY_DATA_ARRAY_DEFAULT_NOTEXISTING_INDEX)
		{
			if ((onlyPolicyOfServer == NULL || _wcsicmp(policies[i].policyPrimaryServer.c_str(), onlyPolicyOfServer) == 0))
			{
				rootItem->children.push_back(AttachToTree(policies, NULL, i, onlyPolicyOfServer, primaryServer));
			}
			else
			{
				for (size_t j = 0; j < policies.size(); ++j)
				{
					if (policies[j].nParentIndex == i && _wcsicmp(policies[j].policyPrimaryServer.c_str(), onlyPolicyOfServer) == 0)
					{
						rootItem->children.push_back(AttachToTree(policies, NULL, i, onlyPolicyOfServer, primaryServer));
						break;
					}
				}
			}
		}
	}
}
