/**
 * @file Command_Help.h
 * @author Copyright (c) 2013 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "Command_Help.h"

#include <Common/era_api/err_codes.h>
#include "../util/stringUtils.h"
#include "../Element.h"
#include "../EracmdExceptions.h"
#include "../Token.h"


Command_Help::Command_Help(CommandList &commandList):commandList(commandList)
{
	possibleValues = LoadStringFromResource(IDS_POSSIBLE_VALUES);
	defaultValue = LoadStringFromResource(IDS_DEFAULT_VALUE);
}


void Command_Help::Execute(elementArray &elements, OutputHandler &out)
{
	SHORT screenX = out.GetScreenWidth();

	std::wostringstream outStream;
	if (arguments.empty())
	{
		outStream << LoadStringFromFileInResource(IDR_HELP_TEXT, L"TEXT") << std::endl;
		std::wstring connectionList = LoadStringFromResource(IDS_HELP_CONNECTION);
		ItemAndDescriptionFormatter frmt;
		while (!connectionList.empty())
		{
			std::wstring wholeString = GetFirstSubstringBeforeDelimiter(connectionList, L"\n", true);
			BuildParamList(frmt, wholeString);
		}
		outStream << frmt.FormatOutput(screenX-1);
	}
	else
	if (arguments.size() == 1 && _wcsicmp(GetArgumentAsString(0).c_str(), L"flags") == 0) // FLAGS
	{
		outStream << LoadStringFromResource(IDS_HELP_FLAGS_HEADER) << std::endl << std::endl;
		std::wstring flagList = LoadStringFromResource(IDS_HELP_FLAGS_FLAGS);
		ItemAndDescriptionFormatter frmt;
		while (!flagList.empty())
		{
			std::wstring wholeString = GetFirstSubstringBeforeDelimiter(flagList, L"\n", true);
			BuildParamList(frmt, wholeString);
		}
		outStream << frmt.FormatOutput(screenX-1);
	}
	else
	if (arguments.size() == 1 && _wcsicmp(GetArgumentAsString(0).c_str(), L"commands") == 0) // COMMANDS
	{
		outStream << LoadStringFromResource(IDS_HELP_COMMANDS) << std::endl << std::endl;

		std::vector<std::wstring> commandStrings;
		commandList.GetCommandStrings(commandStrings);

		ItemAndDescriptionFormatter frmt;
		for (size_t i = 0; i < commandStrings.size(); ++i)
		{
			std::shared_ptr<Command> command = commandList.GetCommand(commandStrings[i]);
			if (command == NULL)
				throw std::exception(std::string("unknown command for help'" + WstringToString(commandStrings[i]) + "'").c_str());
			unsigned int shortId = command->GetHelpShortId();
			if (shortId == 0)
				throw std::exception(std::string("Command '" + WstringToString(commandStrings[i]) + "'does not have help implemented").c_str());

			frmt.Add(commandStrings[i], LoadStringFromResource(shortId));
		}
		frmt.SortByItem();
		outStream << frmt.FormatOutput(screenX-1);
	}
#ifdef _DEBUG
	else
	if (arguments.size() == 1 && (_wcsicmp(GetArgumentAsString(0).c_str(), L"flagshtml") == 0 || _wcsicmp(GetArgumentAsString(0).c_str(), L"commandshtml") == 0)) // HTML
	{
		std::function<std::wstring(const std::wstring &)> toHtml = [](const std::wstring &orig) -> std::wstring
		{
			std::wostringstream ss;

			for (std::wstring::const_iterator it = orig.begin(); it != orig.end(); ++it)
			{
				wchar_t c = *it;

				switch (c)
				{
					case L'<': ss << L"&lt;"; break;
					case L'>': ss << L"&gt;"; break;
					case L'&': ss << L"&amp;"; break;
					case L'\'': ss << L"&apos;"; break;
					case L'"': ss << L"&quot;"; break;
					default: ss << c;
				}
			}

			return ss.str();
		};

		outStream << L"<head><style>table {border-collapse:collapse; } table td {border: 1px solid #555; vertical-align: top; padding: 5px; }</style></head><body><table>";

		if (_wcsicmp(GetArgumentAsString(0).c_str(), L"flagshtml") == 0)
		{
			// header
			outStream << L"<tr>";
				outStream << L"<th>" << toHtml(L"Flag") << L"</th>";
				outStream << L"<th>" << toHtml(L"Description") << L"</th>";
				outStream << L"<th>" << toHtml(L"Possible Values") << L"</th>";
				outStream << L"<th>" << toHtml(L"Default Value") << L"</th>";
				outStream << L"<th>" << toHtml(L"") << L"</th>";
			outStream << L"</tr>";

			std::wstring flagList = LoadStringFromResource(IDS_HELP_FLAGS_FLAGS);

			while (!flagList.empty())
			{
				std::wstring wholeString = GetFirstSubstringBeforeDelimiter(flagList, L"\n", true);

				outStream << L"<tr>";

				std::vector<std::wstring> words;
				SplitToWords(wholeString, L'~', words, true);

				if (words.size() < 2)
					throw std::exception(("missing ~ in param help: " + WstringToString(wholeString)).c_str());

				std::wstring paramName = words[0];
				std::wstring paramHelp = (words.size() > 1) ? words[1] : L"";
				std::wstring paramValues = (words.size() > 2) ? words[2] : L"";
				std::wstring paramDefault = (words.size() > 3) ? words[3] : L"";

				// Flag
				outStream << L"<td>";
				outStream << toHtml(paramName);
				outStream << L"</td>";

				// Description
				outStream << L"<td>";
				outStream << toHtml(paramHelp);
				outStream << L"</td>";

				// Possible Values
				outStream << L"<td>";
				{
					std::vector<std::wstring> values;

					size_t currentValueBegin = 0;
					uint32_t inParentheses = 0;
					for (size_t i = 0; i < paramValues.size(); ++i)
					{
						wchar_t c = paramValues[i];

						if (c == L'(')
							++inParentheses;
						else if (c == L')')
							--inParentheses;
						else if (c == L',' && (inParentheses == 0))
						{
							if (currentValueBegin != i)
							{
								values.push_back(paramValues.substr(currentValueBegin, i - currentValueBegin));
							}
							currentValueBegin = i + 1;
						}
						else if (c == L' ' && (i == currentValueBegin))
						{
							currentValueBegin = i + 1;
						}
					}
					if (currentValueBegin != paramValues.size())
					{
						values.push_back(paramValues.substr(currentValueBegin, paramValues.size() - currentValueBegin));
					}

					outStream << L"<ul>";

					for (std::vector<std::wstring>::const_iterator it = values.begin(); it != values.end(); ++it)
					{
						outStream << L"<li>";
						outStream << toHtml(*it);
						outStream << L"</li>";
					}

					outStream << L"</ul>";
				}

				// Default Value
				outStream << L"<td>";
				outStream << toHtml(paramDefault);
				outStream << L"</td>";

				outStream << L"<td>";
				outStream << L"</td>";

				outStream << L"</tr>";
			}

		}

		if (_wcsicmp(GetArgumentAsString(0).c_str(), L"commandshtml") == 0)
		{
			// header
			outStream << L"<tr>";
				outStream << L"<th>" << toHtml(L"Command") << L"</th>";
				outStream << L"<th>" << toHtml(L"Description") << L"</th>";
				outStream << L"<th>" << toHtml(L"Syntax and Arguments") << L"</th>";
				outStream << L"<th>" << toHtml(L"Examples") << L"</th>";
				outStream << L"<th>" << toHtml(L"") << L"</th>";
			outStream << L"</tr>";

			std::vector<std::wstring> commandStrings;
			commandList.GetCommandStrings(commandStrings);
			for (size_t i_command = 0; i_command < commandStrings.size(); ++i_command)
			{
				const std::wstring &commandString = commandStrings[i_command];
				std::shared_ptr<Command> command = commandList.GetCommand(commandString);
				if (command == NULL)
					throw std::exception(std::string("unknown command for help'" + WstringToString(commandString) + "'").c_str());

				outStream << L"<tr>";

				// Command
				outStream << L"<td>";
				outStream << toHtml(commandString);
				outStream << L"</td>";

				// Description
				outStream << L"<td>";
				{
					unsigned int helpId = command->GetHelpFullId();
					if (helpId == 0)
						throw std::exception(std::string("Command '" + WstringToString(commandString) + "'does not have help implemented").c_str());

					outStream << toHtml(LoadStringFromResource(helpId));
				}
				outStream << L"</td>";

				// Syntax and Arguments
				outStream << L"<td>";
					outStream << toHtml(commandString);

					std::wostringstream ss_desc;
					std::set<unsigned int> printedHelpIds;

					bool inOptional = false;
					for (size_t i_syntax = 0; i_syntax < command->GetSyntax().size(); ++i_syntax)
					{
						const CommandSyntax &arg = command->GetSyntax()[i_syntax];

						if (inOptional && arg.syntaxType != conditional)
						{
							outStream << toHtml(L"]");
							inOptional = false;
						}

						outStream << toHtml(L" ");
						if (arg.syntaxType == optional)
						{
							outStream << toHtml(L"[");
							inOptional = true;
						}

						if (!arg.text.empty()) // keyword
							outStream << toHtml(arg.text);
						else
						{
							std::wstring paramName;
							if (arg.helpId == 0)
								throw std::exception(("missing argument help for command " + WstringToString(commandString)).c_str());
							if (printedHelpIds.find(arg.helpId) == printedHelpIds.end())
							{
								printedHelpIds.insert(arg.helpId);
								std::wstring paramString = LoadStringFromResource(arg.helpId);

								std::vector<std::wstring> words;
								SplitToWords(paramString, L'~', words, true);

								if (words.size() < 2)
									throw std::exception(("missing ~ in param help: " + WstringToString(paramString)).c_str());

								paramName = words[0];
								std::wstring paramHelp = (words.size() > 1) ? words[1] : L"";
								std::wstring paramValues = (words.size() > 2) ? words[2] : L"";
								std::wstring paramDefault = (words.size() > 3) ? words[3] : L"";

								ss_desc << L"<li>";
								ss_desc << L"<strong>" << toHtml(paramName) << L"</strong>";
								ss_desc << L" " << toHtml(paramHelp);
								if (!paramValues.empty())
									ss_desc << L" " << L"<strong>" << toHtml(L"Possible values: ") << L"</strong> " << paramValues;

								if (!paramDefault.empty())
									ss_desc << L" " << L"<strong>" << toHtml(L"Default value: ") << L"</strong> " << paramDefault;
								ss_desc << L"</li>";
							}
							else
							{
								std::wstring paramString = LoadStringFromResource(arg.helpId);
								paramName = GetFirstSubstringBeforeDelimiter(paramString, L"~", false);
							}
							outStream << toHtml(L"<" + paramName + L">");
						}
					}

					if (inOptional)
						outStream << L"]";

					outStream << L"<ul>" << ss_desc.str() << L"</ul>";
				outStream << L"</td>";

				// Examples
				outStream << L"<td>";
				if (command->GetHelpExampleId())
				{
					std::wstring examplesStr = LoadStringFromResource(command->GetHelpExampleId());

					std::vector<std::wstring> examples;
					SplitToWords(examplesStr, L'\n', examples, false);

					for (std::vector<std::wstring>::const_iterator it = examples.begin(); it != examples.end(); ++it)
					{
						outStream << toHtml(*it);
						outStream << L"<br />";
					}
				}
				outStream << L"</td>";

				outStream << L"<td>";
				outStream << L"</td>";

				outStream << L"</tr>";
			}
		}

		outStream << L"</table></body>";
	}
#endif
	else
	if (arguments.size() == 1 && _wcsicmp(GetArgumentAsString(0).c_str(), L"commandlist") == 0) // COMMANDLIST
	{
		std::vector<std::wstring> commandStrings;
		commandList.GetCommandStrings(commandStrings);
		for (size_t i = 0; i < commandStrings.size(); ++i)
		{
			outStream << L"help " << commandStrings[i] << std::endl;
		}
	}
	else // COMMAND SPECIFIC HELP
	{
		std::wstring commandString = elements[0]->GetString();
		if (elements.size() == 2)
			commandString += L" " + elements[1]->GetString();

		std::shared_ptr<Command> command = commandList.GetCommand(commandString);
		if (command == NULL)
			throw EraCMDException(CMDERROR_HELP_UNKNOWN_COMMAND, commandString.c_str());

		UINT helpId = command->GetHelpFullId();
		if (helpId == 0)
			throw std::exception(std::string("Command '" + WstringToString(commandString) + "'does not have help implemented").c_str());

		// full help
		std::wstring helpString = LoadStringFromResource(helpId);
		outStream << helpString << std::endl << std::endl;

		// syntax
		outStream << GetCommandSyntaxHelp(command, commandString, screenX);

		// example
		if (command->GetHelpExampleId())
		{
			std::wstring examples = LoadStringFromResource(command->GetHelpExampleId());
			if (!examples.empty())
			{
				ReplaceAllSubstrings(examples, L"\n", L"\n  ");
				outStream << std::endl << LoadStringFromResource(IDS_EXAMPLE) << std::endl << L"  " << examples;
			}
		}
		outStream << std::endl;
	}

	out.PrintRaw(outStream.str());
}

std::wstring Command_Help::BuildParamList(ItemAndDescriptionFormatter &frmt, const std::wstring &paramLine)
{
	std::vector<std::wstring> words;
	SplitToWords(paramLine, L'~', words, true);

	if (words.size() < 2)
		throw std::exception(("missing ~ in param help: " + WstringToString(paramLine)).c_str());

	std::wstring paramName = words[0];
	std::wstring paramHelp = (words.size() > 1) ? words[1] : L"";
	std::wstring paramValues = (words.size() > 2) ? words[2] : L"";
	std::wstring paramDefault = (words.size() > 3) ? words[3] : L"";
	if (!paramValues.empty()) paramHelp += std::wstring(L"\n") + possibleValues + paramValues;
	if (!paramDefault.empty()) paramHelp += std::wstring(L"\n") + defaultValue + paramDefault;
	frmt.Add(paramName, paramHelp);
	return paramName;
}

std::wstring Command_Help::GetCommandSyntaxHelp(std::shared_ptr<Command> command, std::wstring &commandString, SHORT screenX)
{
	std::wstringstream result;
	result << LoadStringFromResource(IDS_SYNTAX) << std::endl << L"  " << commandString;

	std::set<unsigned int> printedHelpIds;

	ItemAndDescriptionFormatter frmt;
	bool inOptional = false;
	for (size_t i = 0; i < command->GetSyntax().size(); ++i)
	{
		const CommandSyntax &arg = command->GetSyntax()[i];

		if (inOptional && arg.syntaxType != conditional)
		{
			result << L"]";
			inOptional = false;
		}

		result << L" ";
		if (arg.syntaxType == optional)
		{
			result << L"[";
			inOptional = true;
		}

		if (!arg.text.empty()) // keyword
			result << arg.text;
		else
		{
			std::wstring paramName;
			if (arg.helpId == 0)
				throw std::exception(("missing argument help for command " + WstringToString(commandString)).c_str());
			if (printedHelpIds.find(arg.helpId) == printedHelpIds.end())
			{
				printedHelpIds.insert(arg.helpId);
				std::wstring paramString = LoadStringFromResource(arg.helpId);
				paramName = BuildParamList(frmt, paramString);
			}
			else
			{
				std::wstring paramString = LoadStringFromResource(arg.helpId);
				paramName = GetFirstSubstringBeforeDelimiter(paramString, L"~", false);
			}
			result << L"<" << paramName << L">";
		}
	}
	if (inOptional)
		result << L"]";

	std::wstring paramList = frmt.FormatOutput(screenX-1);
	if (!paramList.empty())
	{
		result << std::endl << std::endl << LoadStringFromResource(IDS_PARAMETERS) << std::endl << paramList;
	}
	return result.str();
}
