/**
 * @file Command_Policy.cpp
 * @author Copyright (c) 2013 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "Command_Policy.h"

#include <Common/PolicyConstants.h>
#include <Common/scope_exit.h>
#include <Common/era_api/err_codes.h>
#include "../util/stringUtils.h"
#include "../EracmdExceptions.h"
#include "../PolicyTree.h"

std::wstring Command_Policy::GetServerName()
{
	Era_ServerInfo *serverInfo;

	uint32_t result_getServerInfo = era_GetServerInfo(&serverInfo);
	if (result_getServerInfo != 0)
	{
		throw EraCMDException_FromApi(result_getServerInfo);
	}
	scope_exit se_serverInfo(std::bind(era_ServerInfo_delete, serverInfo));

	return era_ServerInfo_GetServerName(serverInfo);
}

DWORD Command_Policy::StrToDWORD(const std::wstring &str)
{
	std::wstringstream wss;
	wss << str;
	DWORD dwStr;
	wss >> dwStr;
	return dwStr;
}

void Command_Policy::Execute(elementArray &elements, OutputHandler &out)
{
	std::shared_ptr<Element> treeArg = GetArgument(0);

	Era_PolicyQuery *policyQuery = era_PolicyQuery_new();
	if (treeArg)
		era_PolicyQuery_SetDisplayType(policyQuery, ERA_POLICY_DISPLAY_TYPE__TREE);
	else
		era_PolicyQuery_SetDisplayType(policyQuery, ERA_POLICY_DISPLAY_TYPE__TABLE);

	scope_exit se_policyQuery(std::bind(era_PolicyQuery_delete, policyQuery));
	Era_PolicyTableReader *policyTableReader;
	uint32_t result_openQuery = era_OpenPolicyQuery(policyQuery, &policyTableReader);
	if (result_openQuery != 0)
	{
		throw EraCMDException_FromApi(result_openQuery);
	}

	scope_exit se_tableReader(std::bind(era_PolicyTableReader_delete, policyTableReader));
	scope_exit_discard_return<DWORD> se_openQuery(std::bind(era_ClosePolicyQuery, policyTableReader));

	StringVector columns;
	const Era_TableReader *tableReader = era_PolicyTableReader_GetTableReader(policyTableReader);
	uint32_t columnsCount = era_TableReader_GetColumnsCount(tableReader);
	columns.resize(columnsCount);
	for (uint32_t i = 0; i < columnsCount; ++i)
	{
		const Era_ColumnInfo *columnInfo = era_TableReader_GetColumn(tableReader, i);
		columns[i] = era_ColumnInfo_GetHeader(columnInfo);
	}

	StringVector prettyColumns;
	prettyColumns.resize(columnsCount);
	for (uint32_t i = 0; i < columnsCount; ++i)
	{
		const Era_ColumnInfo *columnInfo = era_TableReader_GetColumn(tableReader, i);
		prettyColumns[i] = era_ColumnInfo_GetPrettyHeader(columnInfo);
	}
	std::vector<StringVector> rows;
	StringVector row;

	bool readingComplete = false;
	do
	{
		Era_TableRow *tableRow;
		uint32_t result_readRow = era_ReadPolicyRow(policyTableReader, &tableRow);
		if (result_readRow != 0)
		{
			throw EraCMDException_FromApi(result_readRow);
		}

		if (tableRow != NULL)
		{
			scope_exit se_tableRow(std::bind(era_TableRow_delete, tableRow));

			uint32_t prettyFieldsCount = era_TableRow_GetPrettyFieldsCount(tableRow);
			row.resize(prettyFieldsCount);
			for (uint32_t i = 0; i < prettyFieldsCount; ++i)
			{
				row[i] = era_TableRow_GetPrettyField(tableRow, i);
			}

			rows.push_back(std::move(row));
		}
		else
		{
			readingComplete = true;
		}
	} while (! readingComplete);

	if (treeArg)
	{
		#define N_VIRTUAL_POLICIES 2
		#define N_POLICY_INFO_COLUMNS 8
		#define N_POLICY_TREE_COLUMNS 2

		std::vector<PolicyTreeInfo> policies;
		std::wstring defaultParentPolicyId = std::to_wstring((long long) POLICY_DEFAULT_PARENT_POLICY_ID);
		std::wstring defaultClientPolicyId = std::to_wstring((long long) POLICY_DEFAULT_CLIENT_POLICY_ID);
		policies.reserve(rows.size() + N_VIRTUAL_POLICIES);
		policies.emplace_back(PolicyTreeInfo(POLICY_DEFAULT_PARENT_POLICY_ID, defaultParentPolicyId, LoadStringFromResource(IDS_POLICY_TYPE2), GetServerName(), POLICY_FLAG_DEFAULT_PARENT, POLICY_DEFAULT_PARENT_POLICY_ID, POLICY_DEFAULT_NONE_POLICY_ID, L"", L""));
		policies.emplace_back(PolicyTreeInfo(POLICY_DEFAULT_CLIENT_POLICY_ID, defaultClientPolicyId, LoadStringFromResource(IDS_POLICY_TYPE3), GetServerName(), POLICY_FLAG_VIRTUAL_PARENT_FOR_CLIENTS, POLICY_DEFAULT_CLIENT_POLICY_ID, POLICY_DEFAULT_NONE_POLICY_ID, L"", L""));

		std::for_each(rows.begin(), rows.end(), [&policies](const StringVector &row)
		{
			assert(row.size() >= N_POLICY_INFO_COLUMNS);
			policies.emplace_back(PolicyTreeInfo(StrToDWORD(row[0]), row[0], row[1], row[2], StrToDWORD(row[3]), StrToDWORD(row[4]), StrToDWORD(row[5]), row[6], row[7]));
		});

		assert(columns.size() >= N_POLICY_INFO_COLUMNS);
		columns.resize(N_POLICY_TREE_COLUMNS);
		prettyColumns.resize(N_POLICY_TREE_COLUMNS);
		rows.clear();

		PolicyTreeInfo pbRoot;
		PolicyTreePtr rootItem(new PolicyTree(pbRoot));

		bool bHideForeignPolicies = false;
		std::wstring serverName = GetServerName();
		PolicyTreeUtils::BuildPolicyTree(policies, rootItem, bHideForeignPolicies ? serverName.c_str() : NULL, serverName.c_str());
		PolicyTreeUtils::PrintPolicyTree(rootItem, L"", rows, false);
	}

	TableReader_Const tableReader2(columns, prettyColumns, rows);
	out.Print(tableReader2);
}
