/**
 * @file Command_Report.cpp
 * @author Copyright (c) 2013 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "Command_Report.h"

#include <Common/era_api/err_codes.h>
#include <Common/scope_exit.h>
#include "../EracmdExceptions.h"

Command_Report::Command_Report(CommandFlagsStack &commandFlagsStack)
:
	commandFlagsStack(commandFlagsStack)
{
}

void Command_Report::Execute(elementArray &elements, OutputHandler &out)
{
	Era_ReportType reportType;
	std::wstring typeStr = GetArgumentAsString(0);
	if (_wcsicmp(typeStr.c_str(), L"static") == 0)
		reportType = ERA_REPORT_TYPE__STATIC;
	else if (_wcsicmp(typeStr.c_str(), L"dashboard") == 0)
		reportType = ERA_REPORT_TYPE__DASHBOARD;
	else if (_wcsicmp(typeStr.c_str(), L"generated") == 0)
		reportType = ERA_REPORT_TYPE__GENERATED;
	else
		throw EraCMDException(CMDERROR_REPORT_UNEXPECTED_ARGUMENT, typeStr.c_str());

	Era_ReportQuery *reportQuery = era_ReportQuery_new();
	scope_exit se_reportQuery(std::bind(era_ReportQuery_delete, reportQuery));
	era_ReportQuery_SetType(reportQuery, reportType);
	Era_TableReader *tableReader;
	uint32_t result_openQuery = era_OpenReportQuery(reportQuery, &tableReader);
	if (result_openQuery != 0)
	{
		throw EraCMDException_FromApi(result_openQuery);
	}

	scope_exit se_tableReader(std::bind(era_TableReader_delete, tableReader));
	scope_exit_discard_return<DWORD> se_openQuery(std::bind(era_CloseReportQuery, reportType, tableReader));

	int sortColumn = -1;
	StringVector columns;
	uint32_t columnsCount = era_TableReader_GetColumnsCount(tableReader);
	columns.resize(columnsCount);
	for (uint32_t i = 0; i < columnsCount; ++i)
	{
		const Era_ColumnInfo *columnInfo = era_TableReader_GetColumn(tableReader, i);
		columns[i] = era_ColumnInfo_GetHeader(columnInfo);
		if (sortColumn == -1 && reportType == ERA_REPORT_TYPE__GENERATED && columns[i] == L"date")
			sortColumn = i;
	}

	StringVector prettyColumns;
	prettyColumns.resize(columnsCount);
	for (uint32_t i = 0; i < columnsCount; ++i)
	{
		const Era_ColumnInfo *columnInfo = era_TableReader_GetColumn(tableReader, i);
		prettyColumns[i] = era_ColumnInfo_GetPrettyHeader(columnInfo);
	}
	std::vector<StringVector> rows;
	StringVector row;

	bool readingComplete = false;
	do
	{
		Era_TableRow *tableRow;
		uint32_t result_readRow = era_ReadReportRow(reportType, tableReader, &tableRow);
		if (result_readRow != 0)
		{
			throw EraCMDException_FromApi(result_readRow);
		}

		if (tableRow != NULL)
		{
			scope_exit se_tableRow(std::bind(era_TableRow_delete, tableRow));

			if (commandFlagsStack.CurrentFlags().fieldFormatting == CommandFlags::FIELD_PRETTY)
			{
				uint32_t prettyFieldsCount = era_TableRow_GetPrettyFieldsCount(tableRow);
				row.resize(prettyFieldsCount);
				for (uint32_t i = 0; i < prettyFieldsCount; ++i)
				{
					row[i] = era_TableRow_GetPrettyField(tableRow, i);
				}
			}
			else
			{
				uint32_t fieldsCount = era_TableRow_GetFieldsCount(tableRow);
				row.resize(fieldsCount);
				for (uint32_t i = 0; i < fieldsCount; ++i)
				{
					row[i] = era_TableRow_GetField(tableRow, i);
				}
			}

			rows.push_back(std::move(row));
		}
		else
		{
			readingComplete = true;
		}
	} while (! readingComplete);

	if (sortColumn != -1)
	{
		std::sort(rows.begin(), rows.end(), [sortColumn](StringVector &v1, StringVector &v2) -> bool
		{
			return (_wcsicmp(v1[sortColumn].c_str(), v2[sortColumn].c_str()) > 0);
		});
	}

	TableReader_Const tableReader2(columns, prettyColumns, rows);
	out.Print(tableReader2);
}

