/**
 * @file OutputHandler.h
 * @author Copyright (c) 2013 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "OutputHandler.h"

#include <Common/era_api/err_codes.h>
#include <Common/scope_exit.h>
#include "../console_colors/ConsoleColors_Win.h"
#include "../flags/CommandFlags.h"
#include "../util/stringUtils.h"
#include "../EracmdExceptions.h"
#include "../resource.h"
#include "OutputFormatter.h"
#include "OutputFormatter_Table.h"
#include "OutputFormatter_CSV.h"
#include "PagingStream.h"

// http://stackoverflow.com/a/20229737
class wcrlf_filebuf : public std::basic_filebuf<wchar_t>
{

public:
	wcrlf_filebuf(const wchar_t *wszFilespec, std::ios_base::open_mode _Mode = std::ios_base::out)
		:
		bBomWritten(false)
	{
		memset(awch, 0, sizeof awch);
		BASE::open(wszFilespec, _Mode | std::ios_base::binary);
		pubsetbuf(awch, _countof(awch));

		if (_Mode & std::ios_base::app)
		{
			if (BASE::pubseekoff(0, std::ios_base::end, std::ios_base::out) > 0)
				bBomWritten = true;
		}
	}

	virtual int_type overflow(int_type ch = traits_type::eof())
	{
		if (!bBomWritten)
		{
			bBomWritten = true;
			int_type iRet = BASE::overflow(0xfeff);
			if (iRet != traits_type::not_eof(0xfeff)) return iRet;
		}
		if (ch == '\n') {
			int_type iRet = BASE::overflow('\r');
			if (iRet != traits_type::not_eof('\r')) return iRet;
		}
		return BASE::overflow(ch);
	}

private:
	typedef std::basic_filebuf<wchar_t> BASE;
	wchar_t awch[128];
	bool bBomWritten;
};

class wcrlfofstream : public std::wostream
{
	typedef std::wostream BASE;
public:
	wcrlfofstream(const wchar_t *wszFilespec, std::ios_base::open_mode _Mode = std::ios_base::out)
		:
		std::wostream(&buf),
		buf(wszFilespec, _Mode)
	{}

	wcrlf_filebuf* rdbuf()
	{
		return &buf;
	}

	void close()
	{
		buf.close();
	}

private:
	wcrlf_filebuf buf;
};

OutputHandler::OutputHandler(const CommandFlags &commandFlags, ConsoleColors_Win &consoleColors, const TextWidth &width_console, const TextWidth &width_file)
:
	outputStream(NULL),
	width_console(width_console),
	width_file(width_file),
	headerType(commandFlags.headerType),
	pagedOutput(commandFlags.paged),
	tableClip(commandFlags.tableClip)
{
	switch (commandFlags.outputFormatting)
	{
	case CommandFlags::FORMAT_TABLE:
		{
			outputFormatter.reset(new OutputFormatter_Table(consoleColors));
			break;
		}
	case CommandFlags::FORMAT_CSV:
		{
			WCHAR textDelimiter = L'"';
			WCHAR fieldDelimiter;
			if (commandFlags.outputCSVDelimiter != L'')
				fieldDelimiter = commandFlags.outputCSVDelimiter;
			else
			{
				WCHAR buf[2];
				if (::GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_SLIST, buf, sizeof(buf) / sizeof(TCHAR)) > 0)
					fieldDelimiter = buf[0];
				else
					fieldDelimiter = L',';
			}

			outputFormatter.reset(new OutputFormatter_CSV(fieldDelimiter, textDelimiter, consoleColors));
			break;
		}
	default:
		throw std::exception("Unknown output format");
	}

	if (commandFlags.outputFilePath == L"")
	{
		outputStream = &std::wcout;
	}
	else
	{
		std::ios_base::openmode openMode;
		switch (commandFlags.outputMode)
		{
		case CommandFlags::MODE_OVERWRITE:
			openMode = std::ios_base::out;
			break;
		case CommandFlags::MODE_APPEND:
			openMode = std::ios_base::app;
			break;
		default:
			throw std::exception("Unknown output mode");
		}

		switch (commandFlags.outputEncoding)
		{
		case (CommandFlags::ENCODING_ANSI):
			codePage = CP_ACP;
			convert = true;
			break;
		case (CommandFlags::ENCODING_UTF8):
			codePage = CP_UTF8;
			convert = true;
			break;
		case (CommandFlags::ENCODING_UTF16):
			convert = false;
			break;
		default:
			throw std::exception("Unknown output encoding");
		}

		if (convert)
		{
			std::unique_ptr<std::ofstream> os(new std::ofstream(commandFlags.outputFilePath, openMode));

			if (!os->good())
				throw EraCMDException(CMDERROR_FILE_OPEN_ERROR, commandFlags.outputFilePath.c_str());
			if (codePage == CP_UTF8 && (openMode & std::ios_base::out))
			{
				char bomBytes_utf8[3] = {(char)0xEF, (char)0xBB, (char)0xBF};
				os->write(&bomBytes_utf8[0], 3);
			}

			outputStream = os.release();
		}
		else
		{
			std::unique_ptr<std::wostream> wos(new wcrlfofstream(commandFlags.outputFilePath.c_str(), openMode));

			if (!wos->good())
				throw EraCMDException(CMDERROR_FILE_OPEN_ERROR, commandFlags.outputFilePath.c_str());

			outputStream = wos.release();
		}
	}
}

OutputHandler::~OutputHandler()
{
	if (outputStream != NULL && outputStream != &std::wcout)
		delete outputStream;
}

void OutputHandler::Print(TableReader &reader)
{
	if (convert && (outputStream != &std::wcout))
	{
		std::wostringstream oStream;
		outputFormatter->Print(reader, oStream, headerType, 0, width_file);
		std::vector<char> bytes;
		ConvertStr(oStream.str(), codePage, bytes);

		std::ostream *os = dynamic_cast<std::ostream *>(outputStream);
		if (bytes.size() > 0)
			os->write(&bytes[0], bytes.size());
	}
	else
	{
		StreamModifier_Paging paging(*dynamic_cast<std::wostream *>(outputStream), pagedOutput, []{ std::wcout << LoadStringFromResource(IDS_PRESS_ENTER_FOR_NEXT_PAGE); });

		std::wostream *wos = dynamic_cast<std::wostream *>(outputStream);
		outputFormatter->Print(reader, *wos, headerType, tableClip ? GetScreenWidth() : 0, (outputStream != &std::wcout) ? width_file : width_console);
	}
}

void OutputHandler::PrintRaw(const std::wstring &str)
{
	if (convert && outputStream != &std::wcout)
	{
		std::vector<char> bytes;
		ConvertStr(str, codePage, bytes);

		std::ostream *os = dynamic_cast<std::ostream *>(outputStream);
		if (bytes.size() > 0)
			os->write(&bytes[0], bytes.size());
	}
	else
	{
		StreamModifier_Paging paging(*dynamic_cast<std::wostream *>(outputStream), pagedOutput, []{ std::wcout << LoadStringFromResource(IDS_PRESS_ENTER_FOR_NEXT_PAGE); });

		SHORT screenWidth = GetScreenWidth();
		std::wostream *wos = dynamic_cast<std::wostream *>(outputStream);
		*wos << (screenWidth > 0 ? WrapString(str, screenWidth, false) : str);
	}
}
void OutputHandler::ConvertStr(const std::wstring &str, unsigned int codePage, std::vector<char> &bytes)
{
	if (str.size() == 0) {
		bytes.clear();
		return;
	}

	int bytesLen = ::WideCharToMultiByte(codePage, 0, str.c_str(), (int) str.length(), NULL, 0, NULL, NULL);
	if (bytesLen == 0)
	{
		throw EraCMDException(CMDERROR_ENCODING_CONVERSION_1);
	}
	bytes.resize(bytesLen);
	if (!::WideCharToMultiByte(codePage, 0, str.c_str(), (int) str.length(), &bytes[0], bytesLen, NULL, NULL))
	{
		throw EraCMDException(CMDERROR_ENCODING_CONVERSION_2);
	}
}

SHORT OutputHandler::GetScreenWidth() const
{
	if (outputStream != &std::wcout)
		return 0;
	else
	{
		HANDLE hOut = GetStdHandle(STD_OUTPUT_HANDLE);
		if (hOut == INVALID_HANDLE_VALUE)
			return 0;
		CONSOLE_SCREEN_BUFFER_INFO SBInfo;
		BOOL bRet = GetConsoleScreenBufferInfo(hOut, &SBInfo);
		if (bRet == 0)
			return 0;
		return SBInfo.dwSize.X;
	}
}

