/**
 * @file TextWidth_Console.cpp
 * @author Copyright (c) 2013 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "TextWidth_Console.h"

#include "../resource.h"

typedef BOOL (WINAPI *funcGetConsoleScreenBufferInfoEx) (HANDLE, BOOL, PCONSOLE_FONT_INFOEX);

TextWidth_Console::TextWidth_Console()
{
	// LoadTableFromFile();
	LoadTableFromResources();

	// EnsureTableHasCurrentFontInfo();
	ExtractCurrentFontWidths();
	// SaveTableToFile();
}

void TextWidth_Console::LoadTableFromResources()
{
	HINSTANCE handle = ::GetModuleHandle(NULL);
	HRSRC hRes = ::FindResource(handle, MAKEINTRESOURCE(IDR_FONT_WIDTH_TABLE), RT_RCDATA);
	DWORD dwSize = ::SizeofResource(handle, hRes);
	HGLOBAL hGlob = ::LoadResource(handle, hRes);
	void *pData = ::LockResource(hGlob);

	std::string data((char *)pData, dwSize);

	std::istringstream is(data);
	LoadTableFromStream(is);
}

std::wstring TextWidth_Console::GetCurrentFont()
{
	HINSTANCE hDLL = LoadLibrary(L"kernel32.dll");
	if (! hDLL)
	{
		return L"";
	}

	funcGetConsoleScreenBufferInfoEx GetFunc = (funcGetConsoleScreenBufferInfoEx) ::GetProcAddress(hDLL, "GetCurrentConsoleFontEx");
	if (! GetFunc)
	{
		return L"";
	}

	CONSOLE_FONT_INFOEX fontInfo;
	::memset(&fontInfo, 0, sizeof(fontInfo));
	fontInfo.cbSize = sizeof(fontInfo);

	int res = GetFunc(::GetStdHandle(STD_OUTPUT_HANDLE), FALSE, &fontInfo);
	if (res == 0)
	{
		return L"";
	}
	else
	{
		const bool isTerminal_variableWidth = (fontInfo.FontFamily == 0);

		if (isTerminal_variableWidth)
		{
			return L"Terminal";
		}
		else
		{
			const std::wstring faceName = fontInfo.FaceName;

			const bool isTerminal_fixedWidth = (faceName == L"Terminal");
			if (isTerminal_fixedWidth)
				return L"";
			else
				return faceName;
		}
	}
}

void TextWidth_Console::ExtractCurrentFontWidths()
{
	for (uint32_t i = 0; i < charsCount; ++i)
	{
		charWidths[i] = 1;
	}

	std::wstring fontName = GetCurrentFont();

	FontWidthTable::WidthsForFonts::const_iterator it_font = table.widthsForFonts.find(fontName);
	if (it_font == table.widthsForFonts.end())
	{
		return;
	}

	for (FontWidthTable::Sections::const_iterator it_section = it_font->second.begin(); it_section != it_font->second.end(); ++it_section)
	{
		for (uint32_t i = it_section->range_first; i <= it_section->range_last; ++i)
		{
			charWidths[i] = it_section->width;
		}
	}
}

void TextWidth_Console::LoadTableFromStream(std::istream &is)
{
	table.widthsForFonts.clear();

	uint32_t fontsCount;
	is.read((char *)&fontsCount, sizeof(fontsCount));

	for (uint32_t i_font = 0; i_font < fontsCount; ++i_font)
	{
		uint32_t fontNameLength;
		is.read((char *)&fontNameLength, sizeof(fontNameLength));

		std::vector<wchar_t> fontName;
		fontName.resize(fontNameLength);
		if (fontNameLength > 0)
		{
			is.read((char *)&fontName[0], fontNameLength * sizeof(wchar_t));
		}

		std::wstring fontNameStr = (fontNameLength > 0) ? std::wstring(&fontName[0], fontNameLength) : L"";

		uint32_t sectionsCount;
		is.read((char *)&sectionsCount, sizeof(sectionsCount));

		FontWidthTable::Sections &sections = table.widthsForFonts[fontNameStr];
		for (uint32_t i_section = 0; i_section < sectionsCount; ++i_section)
		{
			FontWidthSection section;

			is.read((char *)&section.range_first, sizeof(section.range_first));
			is.read((char *)&section.range_last, sizeof(section.range_last));
			is.read((char *)&section.width, sizeof(section.width));

			sections.push_back(std::move(section));
		}
	}
}

#ifdef _DEBUG

void TextWidth_Console::LoadTableFromFile()
{
	std::ifstream fs("fontWidthTable.bin", std::ios::binary | std::ios::in);
	LoadTableFromStream(fs);
}

void TextWidth_Console::SaveTableToFile()
{
	std::ofstream fs("fontWidthTable.bin", std::ios::binary | std::ios::out);

	uint32_t fontsCount = (uint32_t) table.widthsForFonts.size();
	fs.write((const char *)&fontsCount, sizeof(fontsCount));

	for (FontWidthTable::WidthsForFonts::const_iterator it_font = table.widthsForFonts.begin(); it_font != table.widthsForFonts.end(); ++it_font)
	{
		const std::wstring &fontName = it_font->first;
		const FontWidthTable::Sections &sections = it_font->second;

		uint32_t fontNameLength = (uint32_t) fontName.size();
		fs.write((const char *)&fontNameLength, sizeof(fontNameLength));
		fs.write((const char *)fontName.c_str(), fontNameLength * sizeof(wchar_t));

		uint32_t sectionsCount = (uint32_t) sections.size();
		fs.write((const char *)&sectionsCount, sizeof(sectionsCount));

		for (FontWidthTable::Sections::const_iterator it_section = sections.begin(); it_section != sections.end(); ++it_section)
		{
			const FontWidthSection &section = *it_section;

			fs.write((const char *)&section.range_first, sizeof(section.range_first));
			fs.write((const char *)&section.range_last, sizeof(section.range_last));
			fs.write((const char *)&section.width, sizeof(section.width));
		}
	}
}

void TextWidth_Console::EnsureTableHasCurrentFontInfo()
{
	std::wstring fontName = GetCurrentFont();

	if (table.widthsForFonts.find(fontName) != table.widthsForFonts.end())
	{
		return;
	}

	FontWidthTable::Sections &sections = table.widthsForFonts[fontName];

	CONSOLE_SCREEN_BUFFER_INFO info;
	::GetConsoleScreenBufferInfo(GetStdHandle(STD_OUTPUT_HANDLE), &info);
	COORD origPos = info.dwCursorPosition;

	uint32_t group_first = 0;
	int group_cells = 1;
	COORD testPos;
	testPos.X = 0;
	testPos.Y = 0;

	for (uint32_t i = 0; i < charsCount; ++i)
	{
		::SetConsoleCursorPosition(GetStdHandle(STD_OUTPUT_HANDLE), testPos);

		wchar_t c = (wchar_t)i;
		if (c == L'\t')
		{
			c = L' ';
		}

		std::wcout << c;

		CONSOLE_SCREEN_BUFFER_INFO info;
		::GetConsoleScreenBufferInfo(GetStdHandle(STD_OUTPUT_HANDLE), &info);

		uint32_t current_cells = (info.dwCursorPosition.X);

		if (current_cells != group_cells)
		{
			if (group_cells != 1)
			{
				sections.push_back(FontWidthSection(group_first, i - 1, group_cells));
			}

			group_cells = current_cells;
			group_first = i;
		}
	}

	if (group_cells != 1)
	{
		sections.push_back(FontWidthSection(group_first, charsCount - 1, group_cells));
	}

	::SetConsoleCursorPosition(GetStdHandle(STD_OUTPUT_HANDLE), origPos);
}

#endif

size_t TextWidth_Console::GetWidth(const std::wstring &str) const
{
	size_t width = 0;

	for (std::wstring::const_iterator it = str.begin(); it != str.end(); ++it)
	{
		width += charWidths[*it];
	}

	return width;
}

std::wstring TextWidth_Console::AlignTo(const std::wstring &str, const size_t width) const
{
	std::wstring result;
	size_t currentWidth = 0;

	for (std::wstring::const_iterator it = str.begin(); it != str.end(); ++it)
	{
		if (currentWidth + charWidths[*it] > width)
		{
			break;
		}

		result.push_back(*it);
		currentWidth += charWidths[*it];
	}

	result += std::wstring(width - currentWidth, L' ');

	return result;
}

